"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = __importDefault(require("axios"));
const fs_1 = __importDefault(require("fs"));
const react_native_fs_1 = require("react-native-fs");
const constants_1 = require("../constants");
const remote_oca_1 = require("../legacy/resolver/remote-oca");
const bundleFileName = 'bundle.json';
const ocaPath = `${__dirname}/fixtures/${constants_1.ocaCacheDataFileName}`;
const ocaAsString = fs_1.default.readFileSync(ocaPath, 'utf8');
const indexPath = `${__dirname}/fixtures/${constants_1.defaultBundleIndexFileName}`;
const indexAsString = fs_1.default.readFileSync(indexPath, 'utf8');
const bundlePath = `${__dirname}/fixtures/${bundleFileName}`;
const bundleAsString = fs_1.default.readFileSync(bundlePath, 'utf8');
const initialEtag = '3379dc5c38ac34ab';
const changedEtag = '4c0afc9bfdb3bf6b';
const keysToUpdated = [
    '79e343ab4362d93974cdf78b00a916a5b3659178dfd170fcff90f1b5f10ceb72',
    '29bf8f8729f235450980d0b273f9ca49b7ab73b7358467bf52b311bf40464bb0',
];
// Testing person credential as an example
const personSha = '9187f651eab59811d8ec1c1257628c08f6daa1eb520bf379f022425f3a9701cf';
const qualifiedCredDefId = 'did:indy:candy:RGjWbW1eycP7FrMf4QJvX8/anoncreds/v0/CLAIM_DEF/13/Person';
const qualifiedSchemaId = 'did:indy:candy:RGjWbW1eycP7FrMf4QJvX8/anoncreds/v0/SCHEMA/Person/1.0';
const getMock = jest.fn().mockImplementation((url) => {
    const fileName = url.split('/').pop();
    switch (fileName) {
        case 'ocabundles.json':
            return Promise.resolve({
                status: 200,
                data: JSON.parse(indexAsString),
                headers: { etag: initialEtag },
            });
        case 'OCABundle.json':
            return Promise.resolve({
                status: 200,
                data: JSON.parse(bundleAsString),
            });
        default:
            return Promise.reject(new Error('Not Found'));
    }
});
jest.mock('axios', () => ({
    create: jest.fn().mockReturnValue({
        get: jest.fn(),
    }),
}));
jest.useFakeTimers({ legacyFakeTimers: true });
jest.mock('react-native-fs', () => ({
    exists: jest.fn().mockResolvedValue(true),
    mkdir: jest.fn().mockResolvedValue(true),
    readFile: jest.fn().mockImplementation((filePath) => {
        const fileName = filePath.split('/').pop();
        switch (fileName) {
            case constants_1.ocaCacheDataFileName: {
                return Promise.resolve(ocaAsString);
            }
            case constants_1.defaultBundleIndexFileName:
                return Promise.resolve(indexAsString);
            case personSha:
                return Promise.resolve(bundleAsString);
            default:
                return Promise.resolve('{}');
        }
    }),
    writeFile: jest.fn().mockResolvedValue(true),
    unlink: jest.fn().mockResolvedValue(true),
}));
describe('RemoteOCABundleResolver', () => {
    let resolver;
    beforeEach(() => {
        resolver = new remote_oca_1.RemoteOCABundleResolver('http://example.com');
        // resolver.log = new MiniLogger()
        jest.clearAllMocks();
        // @ts-expect-error some weird type
        axios_1.default.create().get.mockImplementation(getMock);
    });
    it('should be created', async () => {
        expect(resolver).toBeTruthy();
        expect(resolver).toBeInstanceOf(remote_oca_1.RemoteOCABundleResolver);
        expect(resolver).toHaveProperty('indexFileName', constants_1.defaultBundleIndexFileName);
        expect(resolver).toHaveProperty('cacheDataFileName', constants_1.ocaCacheDataFileName);
        expect(resolver['axiosInstance']).not.toBeNull();
        expect(axios_1.default.create().get).toHaveBeenCalledTimes(0);
        expect(resolver.indexFileEtag).toEqual('');
        expect(resolver['indexFile']).toMatchSnapshot();
    });
    it('should update demand', async () => {
        await resolver.checkForUpdates();
        expect(axios_1.default.create().get).toHaveBeenCalledTimes(16); // 1 index + 15 bundles
        expect(react_native_fs_1.writeFile).toHaveBeenCalledTimes(17); // 1 index + 15 bundles + 1 cache
        expect(react_native_fs_1.readFile).toHaveBeenCalledTimes(1); // 1 cache
        expect(react_native_fs_1.unlink).toHaveBeenCalledTimes(0); // 0 deletes
        expect(react_native_fs_1.exists).toHaveBeenCalledTimes(3); // 1 index + 1 cache + 1 directory
        expect(react_native_fs_1.mkdir).toHaveBeenCalledTimes(0); // mock always return true
        expect(resolver.indexFileEtag).toEqual(initialEtag);
        expect(resolver['indexFile']).toMatchSnapshot();
    });
    it.skip('should check index periodically', async () => {
        await resolver.checkForUpdates();
        await resolver.checkForUpdates();
        await resolver.checkForUpdates();
        expect(axios_1.default.create().get).toHaveBeenCalledTimes(18); // 3 index + 15 bundles
        expect(react_native_fs_1.writeFile).toHaveBeenCalledTimes(17); // 1 index + 15 bundles + 1 cache
        expect(react_native_fs_1.readFile).toHaveBeenCalledTimes(1);
        expect(react_native_fs_1.unlink).toHaveBeenCalledTimes(0);
        expect(react_native_fs_1.exists).toHaveBeenCalledTimes(5);
        expect(react_native_fs_1.mkdir).toHaveBeenCalledTimes(0);
    });
    it('should check fetch updated bundles', async () => {
        await resolver.checkForUpdates();
        expect(axios_1.default.create().get).toHaveBeenCalledTimes(16); // 1 index + 15 bundles
        expect(resolver.indexFileEtag).toEqual(initialEtag);
        const newIndexAsJSON = JSON.parse(indexAsString);
        for (const key of Object.keys(newIndexAsJSON)) {
            if (keysToUpdated.includes(newIndexAsJSON[key].sha256)) {
                newIndexAsJSON[key].sha256 = newIndexAsJSON[key].sha256.slice(0, 7);
            }
        }
        // @ts-expect-error some weird type
        axios_1.default.create().get.mockImplementation((url) => {
            const fileName = url.split('/').pop();
            switch (fileName) {
                case 'ocabundles.json':
                    return Promise.resolve({
                        status: 200,
                        data: newIndexAsJSON,
                        headers: { etag: changedEtag },
                    });
                case 'OCABundle.json':
                    return Promise.resolve({
                        status: 200,
                        data: JSON.parse(bundleAsString),
                    });
                default:
                    return Promise.reject(new Error('Not Found'));
            }
        });
        await resolver.checkForUpdates();
        expect(axios_1.default.create().get).toHaveBeenCalledTimes(19); // 2 index + 17 bundles
        expect(react_native_fs_1.writeFile).toHaveBeenCalledTimes(21); // 2 index + 17 bundles + 2 cache
        expect(react_native_fs_1.readFile).toHaveBeenCalledTimes(1);
        expect(react_native_fs_1.unlink).toHaveBeenCalledTimes(2); // 2 deletes
        expect(react_native_fs_1.exists).toHaveBeenCalledTimes(4); // 2 index + 1 cache + 1 directory
        expect(react_native_fs_1.mkdir).toHaveBeenCalledTimes(0);
        expect(resolver.indexFileEtag).toEqual(changedEtag);
        expect(resolver['indexFile']).toMatchSnapshot();
    });
    it('should check unqualified cred def id as key if given qualified cred def id', async () => {
        await resolver.checkForUpdates();
        const bundle = await resolver.resolve({ identifiers: { credentialDefinitionId: qualifiedCredDefId } });
        expect(bundle).toBeDefined();
    });
    it('should check unqualified schema id as key if given qualified schema id', async () => {
        await resolver.checkForUpdates();
        const bundle = await resolver.resolve({ identifiers: { schemaId: qualifiedSchemaId } });
        expect(bundle).toBeDefined();
    });
});
