"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.RemoteOCABundleResolver = void 0;
const anoncreds_1 = require("@credo-ts/anoncreds");
const axios_1 = __importDefault(require("axios"));
const react_native_fs_1 = require("react-native-fs");
const constants_1 = require("../../constants");
const types_1 = require("../../types");
const utils_1 = require("../../utils");
const oca_1 = require("./oca");
var OCABundleQueueEntryOperation;
(function (OCABundleQueueEntryOperation) {
    OCABundleQueueEntryOperation["Add"] = "add";
    OCABundleQueueEntryOperation["Remove"] = "remove";
})(OCABundleQueueEntryOperation || (OCABundleQueueEntryOperation = {}));
class RemoteOCABundleResolver extends oca_1.DefaultOCABundleResolver {
    constructor(indexFileBaseUrl, options) {
        super({}, options);
        this.indexFile = {};
        this.cacheDataFileName = constants_1.ocaCacheDataFileName;
        /**
         * Loads the cache data from the local storage.
         * @returns A promise that resolves to the cache data file, or undefined if the cache file does not exist or cannot be loaded.
         */
        this.loadCacheData = async () => {
            const cacheFileExists = await this.checkFileExists(this.cacheDataFileName);
            if (!cacheFileExists) {
                return;
            }
            const data = await this.loadFileFromLocalStorage(this.cacheDataFileName);
            if (!data) {
                return;
            }
            const cacheData = JSON.parse(data);
            return cacheData;
        };
        /**
         * Saves the cache data to local storage.
         *
         * @param cacheData - The cache data to be saved.
         * @returns A promise that resolves to a boolean indicating whether the save operation was successful.
         */
        this.saveCacheData = async (cacheData) => {
            const cacheDataAsString = JSON.stringify(cacheData);
            return this.saveFileToLocalStorage(this.cacheDataFileName, cacheDataAsString);
        };
        /**
         * Processes the queue of OCABundleQueueEntry items.
         *
         * @param items - An array of OCABundleQueueEntry items to process.
         * @returns A promise that resolves to an array of processed OCABundleQueueEntry items.
         */
        this.processQueue = async (items) => {
            var _a, _b, _c;
            const processed = Array();
            const operations = [];
            for (const q of items) {
                const hash = q.sha256;
                (_a = this.log) === null || _a === void 0 ? void 0 : _a.info(`Processing op ${q.operation} for ${hash}`);
                switch (q.operation) {
                    case OCABundleQueueEntryOperation.Add:
                        {
                            const path = this.findPathBySha256(hash);
                            if (!path) {
                                continue;
                            }
                            operations.push(this.fetchOCABundle(path));
                        }
                        break;
                    case OCABundleQueueEntryOperation.Remove:
                        operations.push(this.removeFileFromLocalStorage(hash));
                        break;
                }
            }
            try {
                // Check which operations were successful, and remove them from
                // the queue.
                const settled = await Promise.allSettled(operations);
                for (const i in settled) {
                    if (settled[i].status === 'fulfilled') {
                        processed.push(items[i]);
                    }
                }
                return (_b = Array.from(processed)) !== null && _b !== void 0 ? _b : [];
            }
            catch (error) {
                (_c = this.log) === null || _c === void 0 ? void 0 : _c.error(`Failed to process some operations, ${error}`);
                return [];
            }
        };
        /**
         * Prepares the bundle queue based on the new and old index files.
         * It compares the SHA256 hashes of the files in the new and old index files
         * and determines which files should be removed and which files should be added
         * to the bundle queue.
         *
         * @param newIndexFile - The new bundle index file.
         * @param oldIndexFile - The old bundle index file.
         * @returns An array of `OCABundleQueueEntry` objects representing the files to be removed and added.
         */
        this.prepareBundleQueue = (newIndexFile, oldIndexFile) => {
            var _a;
            const oldIndexItemHashes = [...new Set(Object.keys(oldIndexFile).map((key) => oldIndexFile[key].sha256))];
            const newIndexItemHashes = [...new Set(Object.keys(newIndexFile).map((key) => newIndexFile[key].sha256))];
            // if the SHA256 is in the old index file but not in
            // the new index file, it should be removed.
            const hashesToRemove = oldIndexItemHashes
                .filter((hash) => !newIndexItemHashes.includes(hash))
                .map((hash) => ({
                sha256: hash,
                operation: OCABundleQueueEntryOperation.Remove,
            }));
            // if the SHA256 is in the new index file but not in
            // the old index file, it should be added.
            const hashesToAdd = newIndexItemHashes
                .filter((hash) => !oldIndexItemHashes.includes(hash))
                .map((hash) => ({
                sha256: hash,
                operation: OCABundleQueueEntryOperation.Add,
            }));
            (_a = this.log) === null || _a === void 0 ? void 0 : _a.info(`Files to remove ${hashesToRemove.length}, add ${hashesToAdd.length}`);
            return [...hashesToRemove, ...hashesToAdd];
        };
        /**
         * Finds the SHA-256 hash associated with a given path in the index file.
         *
         * @param {string} path - The path to search for in the index file.
         * @returns {string | null} The SHA-256 hash associated with the path if found, or null if not found.
         */
        this.findSha256ByPath = (path) => {
            for (const key in this.indexFile) {
                if (this.indexFile[key].path === path) {
                    return this.indexFile[key].sha256;
                }
            }
            return null;
        };
        /**
         * Finds the path associated with the given SHA256 hash in the index file.
         *
         * @param sha256 - The SHA256 hash to search for.
         * @returns The path associated with the SHA256 hash, or null if not found.
         */
        this.findPathBySha256 = (sha256) => {
            for (const key of Object.keys(this.indexFile)) {
                const hash = this.indexFile[key].sha256;
                if (hash === sha256) {
                    return this.indexFile[key].path;
                }
            }
            return null;
        };
        /**
         * Returns the file name for the bundle at the specified path.
         *
         * @param path - The path of the bundle.
         * @returns The file name for the bundle, or null if not found.
         */
        this.fileNameForBundleAtPath = (path) => {
            return this.findSha256ByPath(path);
        };
        /**
         * Returns the file storage path for the remote OCA resolver.
         * The file storage path is the concatenation of the `CachesDirectoryPath` and `ocaBundleStorageDirectory`.
         *
         * @returns The file storage path.
         */
        this.fileStoragePath = () => {
            return `${react_native_fs_1.CachesDirectoryPath}/${constants_1.ocaBundleStorageDirectory}`;
        };
        /**
         * Checks if a file exists at a specific path in the document directory.
         *
         * @param {string} fileName - The name of the file to check.
         * @returns {Promise<boolean>} A promise that resolves to true if the file exists, or false otherwise.
         * @throws Will throw an error if the existence check fails.
         */
        this.checkFileExists = async (fileName) => {
            var _a, _b;
            const pathToFile = `${this.fileStoragePath()}/${fileName}`;
            try {
                const fileExists = await (0, react_native_fs_1.exists)(pathToFile);
                (_a = this.log) === null || _a === void 0 ? void 0 : _a.info(`File ${fileName} ${fileExists ? 'does' : 'does not'} exist at ${pathToFile}`);
                return fileExists;
            }
            catch (error) {
                (_b = this.log) === null || _b === void 0 ? void 0 : _b.error(`Failed to check existence of ${fileName} at ${pathToFile}`);
            }
            return false;
        };
        /**
         * Creates a working directory if it does not already exist.
         *
         * @returns A promise that resolves to a boolean indicating whether the directory was created successfully.
         */
        this.createWorkingDirectoryIfNotExists = async () => {
            var _a;
            const workSpace = this.fileStoragePath();
            const pathDoesExist = await (0, react_native_fs_1.exists)(workSpace);
            if (!pathDoesExist) {
                try {
                    await (0, react_native_fs_1.mkdir)(workSpace);
                    return true;
                }
                catch (error) {
                    (_a = this.log) === null || _a === void 0 ? void 0 : _a.error(`Failed to create directory ${workSpace}`);
                    return false;
                }
            }
            return true;
        };
        /**
         * Saves a string of data to a file in the local storage.
         *
         * @param {string} fileName - The name of the file to save.
         * @param {string} data - The data to write to the file.
         * @returns {Promise<boolean>} A promise that resolves to true if the file was saved successfully, or false otherwise.
         * @throws Will throw an error if the write operation fails.
         */
        this.saveFileToLocalStorage = async (fileName, data) => {
            var _a, _b;
            const pathToFile = `${this.fileStoragePath()}/${fileName}`;
            try {
                await (0, react_native_fs_1.writeFile)(pathToFile, data, 'utf8');
                (_a = this.log) === null || _a === void 0 ? void 0 : _a.info(`File ${fileName} saved to ${pathToFile}`);
                return true;
            }
            catch (error) {
                (_b = this.log) === null || _b === void 0 ? void 0 : _b.error(`Failed to save file ${fileName} to ${pathToFile}, ${error}`);
            }
            return false;
        };
        /**
         * Loads a file from local storage.
         *
         * @param fileName - The name of the file to load.
         * @returns A promise that resolves to the contents of the file, or undefined if the file does not exist.
         */
        this.loadFileFromLocalStorage = async (fileName) => {
            var _a, _b, _c;
            const pathToFile = `${this.fileStoragePath()}/${fileName}`;
            try {
                const fileExists = await this.checkFileExists(fileName);
                if (!fileExists) {
                    (_a = this.log) === null || _a === void 0 ? void 0 : _a.warn(`Missing ${fileName} from ${pathToFile}`);
                    return;
                }
                const data = await (0, react_native_fs_1.readFile)(pathToFile, 'utf8');
                (_b = this.log) === null || _b === void 0 ? void 0 : _b.info(`File ${fileName} loaded from ${pathToFile}`);
                return data;
            }
            catch (error) {
                (_c = this.log) === null || _c === void 0 ? void 0 : _c.error(`Failed to load file ${fileName} from ${pathToFile}`);
            }
        };
        /**
         * Removes a file from the local storage.
         *
         * @param fileName - The name of the file to be removed.
         * @returns A promise that resolves to a boolean indicating whether the file was successfully removed.
         */
        this.removeFileFromLocalStorage = async (fileName) => {
            var _a;
            const pathToFile = `${this.fileStoragePath()}/${fileName}`;
            try {
                await (0, react_native_fs_1.unlink)(pathToFile);
                return true;
            }
            catch (error) {
                (_a = this.log) === null || _a === void 0 ? void 0 : _a.error(`Failed to unlink file ${fileName} from ${pathToFile}`);
            }
            return false;
        };
        /**
         * Fetches an OCA bundle from a remote resource and saves it to local storage.
         * @param path - The path of the remote resource.
         * @returns A promise that resolves to a boolean indicating whether the fetch and save operation was successful.
         */
        this.fetchOCABundle = async (path) => {
            var _a, _b;
            const response = await this.axiosInstance.get(path);
            const { status } = response;
            if (status !== 200) {
                (_a = this.log) === null || _a === void 0 ? void 0 : _a.error(`Failed to fetch remote resource at ${path}`);
                return false;
            }
            const fileName = this.fileNameForBundleAtPath(path);
            if (!fileName) {
                (_b = this.log) === null || _b === void 0 ? void 0 : _b.error(`Failed to determine file name ${fileName} for save`);
                return false;
            }
            return this.saveFileToLocalStorage(fileName, JSON.stringify(response.data));
        };
        /**
         * Loads the OCA index from a remote location and processes it.
         * If the remote resource is not available, it falls back to the cached index file.
         * If the index file has not changed, it uses the existing data.
         * If the index file has changed, it refreshes the index file and the bundles.
         *
         * @param filePath - The path to the remote OCA index file.
         * @returns A Promise that resolves when the index file and bundles have been processed.
         * @throws If there is an error fetching or processing the OCA index.
         */
        this.loadOCAIndex = async (filePath) => {
            var _a, _b, _c, _d, _e, _f;
            try {
                const response = await this.axiosInstance.get(filePath);
                const { status } = response;
                const { etag } = response.headers;
                if (status !== 200) {
                    (_a = this.log) === null || _a === void 0 ? void 0 : _a.error(`Failed to fetch remote resource at ${filePath}`);
                    // failed to fetch, use the cached index file
                    // if available
                    const data = await this.loadFileFromLocalStorage(filePath);
                    if (data) {
                        (_b = this.log) === null || _b === void 0 ? void 0 : _b.info(`Using index file ${filePath} from cache`);
                        this.indexFile = JSON.parse(data);
                    }
                    return;
                }
                if (etag && etag === this.indexFileEtag) {
                    (_c = this.log) === null || _c === void 0 ? void 0 : _c.info(`Index file ${filePath} has not changed, etag ${etag}`);
                    // etag is the same, no need to refresh
                    this.indexFile = response.data;
                    const fetchMissingFiles = true;
                    const status = await this.verifyCacheIntegrity(fetchMissingFiles);
                    if (!status) {
                        (_d = this.log) === null || _d === void 0 ? void 0 : _d.error(`Cache integrity broken, unable to re-fetch missing files`);
                    }
                    return;
                }
                // etag is different, we need to refresh the
                // index file and the bundles.
                const items = this.prepareBundleQueue(response.data, this.indexFile);
                this.indexFile = response.data;
                this.indexFileEtag = etag;
                const processed = await this.processQueue(items);
                if (processed.length !== items.length) {
                    (_e = this.log) === null || _e === void 0 ? void 0 : _e.info(`Processed ${processed.length} items, expected ${items.length}`);
                }
                await this.saveFileToLocalStorage(filePath, JSON.stringify(this.indexFile));
            }
            catch (error) {
                (_f = this.log) === null || _f === void 0 ? void 0 : _f.error(`Failed to fetch OCA index ${filePath}`);
            }
        };
        /**
         * Loads the OCABundle from the specified path.
         *
         * @param path - The path of the OCABundle.
         * @returns A promise that resolves to an array of IOverlayBundleData.
         */
        this.loadOCABundle = async (path) => {
            var _a;
            // check if the file exists in the local storage
            // if it does, load it from there.
            const fileName = this.fileNameForBundleAtPath(path);
            if (!fileName) {
                (_a = this.log) === null || _a === void 0 ? void 0 : _a.error(`Failed to determine file name ${fileName} for save`);
                return [];
            }
            const cachedData = await this.loadFileFromLocalStorage(fileName);
            if (cachedData) {
                return JSON.parse(cachedData);
            }
            return [];
        };
        /**
         * Finds a matching identifier in the index file based on the provided identifiers.
         * The order of the identifiers matters if more than one candidate exists in the index file.
         *
         * @param identifiers - The identifiers to match against the index file.
         * @returns The matching identifier, or undefined if no match is found.
         */
        this.matchBundleIndexEntry = (identifiers) => {
            const { schemaId, credentialDefinitionId, templateId } = identifiers;
            // also check unqualified schema and cred def id's if qualified versions exist
            const unqualifiedSchemaId = (schemaId === null || schemaId === void 0 ? void 0 : schemaId.startsWith('did:indy:')) ? (0, anoncreds_1.getUnQualifiedDidIndyDid)(schemaId) : undefined;
            const unqualifiedCredDefId = (credentialDefinitionId === null || credentialDefinitionId === void 0 ? void 0 : credentialDefinitionId.startsWith('did:indy:'))
                ? (0, anoncreds_1.getUnQualifiedDidIndyDid)(credentialDefinitionId)
                : undefined;
            // If more than one candidate identifier exists in the index file then
            // order matters here.
            const candidates = [schemaId, unqualifiedSchemaId, credentialDefinitionId, unqualifiedCredDefId, templateId].filter((value) => value !== undefined && value !== null && value !== '');
            if (candidates.length === 0) {
                return undefined;
            }
            const keys = Object.keys(this.indexFile);
            const identifier = candidates.find((c) => keys.includes(c));
            return identifier;
        };
        this.indexFileName = (options === null || options === void 0 ? void 0 : options.indexFileName) || constants_1.defaultBundleIndexFileName;
        this.axiosInstance = axios_1.default.create({
            baseURL: indexFileBaseUrl,
        });
    }
    /**
     * Sets the value of the index file ETag.
     *
     * @param value - The new value for the index file ETag.
     */
    set indexFileEtag(value) {
        if (!value) {
            return;
        }
        this._indexFileEtag = value;
        this.saveCacheData({
            indexFileEtag: value,
            updatedAt: new Date(),
        }).catch((error) => {
            var _a;
            (_a = this.log) === null || _a === void 0 ? void 0 : _a.error(`Failed to save cache data, ${error}`);
        });
    }
    /**
     * Gets the ETag of the index file.
     *
     * @returns The ETag of the index file, or an empty string if not available.
     */
    get indexFileEtag() {
        return this._indexFileEtag || '';
    }
    /**
     * Checks for updates in the OCA (Overlay Capture Architecture) index.
     * If the index file ETag is not available, it loads the cache data and retrieves the ETag from it.
     * Then, it loads the OCA index.
     * @returns A promise that resolves when the update check is complete.
     */
    async checkForUpdates() {
        var _a, _b;
        await this.createWorkingDirectoryIfNotExists();
        if (!this.indexFileEtag) {
            (_a = this.log) === null || _a === void 0 ? void 0 : _a.info('Loading cache data');
            const cacheData = await this.loadCacheData();
            if (cacheData) {
                this.indexFileEtag = cacheData.indexFileEtag;
            }
        }
        (_b = this.log) === null || _b === void 0 ? void 0 : _b.info('Loading OCA index now');
        await this.loadOCAIndex(this.indexFileName);
    }
    /**
     * Resolves the OCABundle based on the given parameters.
     * @param params - The parameters for resolving the OCABundle.
     * @param params.identifiers - The identifiers used to match the OCABundle.
     * @param params.language - The language of the OCABundle (optional).
     * @returns A Promise that resolves to the OCABundle or undefined.
     */
    async resolve(params) {
        var _a, _b;
        const language = params.language || constants_1.defaultBundleLanguage;
        const identifier = this.matchBundleIndexEntry(params.identifiers);
        if (!identifier || !(identifier in this.bundles || identifier in this.indexFile)) {
            return Promise.resolve(undefined);
        }
        if (identifier in this.bundles) {
            return Promise.resolve(new oca_1.OCABundle(this.bundles[identifier], Object.assign(Object.assign({}, this.options), { language: language !== null && language !== void 0 ? language : this.options.language })));
        }
        const bundlePath = (_a = this.indexFile[identifier]) === null || _a === void 0 ? void 0 : _a.path;
        if (!bundlePath) {
            return Promise.resolve(undefined);
        }
        try {
            const bundleData = await this.loadOCABundle(bundlePath);
            const overlayBundle = new types_1.OverlayBundle((_b = params.identifiers.credentialDefinitionId) !== null && _b !== void 0 ? _b : '', bundleData[0]);
            const overlay = overlayBundle.overlays.find((overlay) => overlay.type === oca_1.BrandingOverlayType.Branding01 || overlay.type === oca_1.BrandingOverlayType.Branding10);
            if (!overlay) {
                overlayBundle.overlays.push(...this.getFallbackBrandingOverlays(overlayBundle.credentialDefinitionId, overlayBundle.captureBase.digest));
            }
            this.bundles[identifier] = overlayBundle;
            return new oca_1.OCABundle(overlayBundle, Object.assign(Object.assign({}, this.options), { language: language !== null && language !== void 0 ? language : this.options.language }));
        }
        catch (error) {
            // probably couldn't parse the overlay bundle.
            return Promise.resolve(undefined);
        }
    }
    /**
     * Retrieves the fallback branding overlays for a given credential definition ID and capture base.
     * @param credentialDefinitionId - The ID of the credential definition.
     * @param captureBase - The capture base for the overlays.
     * @returns An array of fallback branding overlays.
     */
    getFallbackBrandingOverlays(credentialDefinitionId, captureBase) {
        const legacyBrandingOverlay = new types_1.LegacyBrandingOverlay(credentialDefinitionId, {
            capture_base: captureBase,
            type: oca_1.BrandingOverlayType.Branding01,
            background_color: (0, utils_1.generateColor)(credentialDefinitionId),
        });
        const brandingOverlay = new types_1.BrandingOverlay(credentialDefinitionId, {
            capture_base: captureBase,
            type: oca_1.BrandingOverlayType.Branding10,
            primary_background_color: (0, utils_1.generateColor)(credentialDefinitionId),
        });
        return [legacyBrandingOverlay, brandingOverlay];
    }
    async verifyCacheIntegrity(fetchMissing = false) {
        var _a, _b, _c, _d, _e, _f;
        const indexItemHashes = [...new Set(Object.keys(this.indexFile).map((key) => this.indexFile[key].sha256))];
        const queue = [];
        (_a = this.log) === null || _a === void 0 ? void 0 : _a.info(`Checking bundle cache integrity`);
        for (const hash of indexItemHashes) {
            const fileName = this.fileNameForBundleAtPath(this.findPathBySha256(hash));
            if (!fileName) {
                continue;
            }
            const fileExists = await this.checkFileExists(fileName);
            if (!fileExists) {
                (_b = this.log) === null || _b === void 0 ? void 0 : _b.warn(`File ${fileName} does not exist, re-fetching`);
                queue.push({
                    sha256: hash,
                    operation: OCABundleQueueEntryOperation.Add,
                });
            }
        }
        if (queue.length === 0) {
            (_c = this.log) === null || _c === void 0 ? void 0 : _c.info(`Cache integrity verified`);
            return true;
        }
        if (queue.length > 0 && !fetchMissing) {
            (_d = this.log) === null || _d === void 0 ? void 0 : _d.info(`Missing ${queue.length} files, cache broken`);
            return false;
        }
        const processed = await this.processQueue(queue);
        if (processed.length !== queue.length) {
            (_e = this.log) === null || _e === void 0 ? void 0 : _e.error(`Processed ${processed.length} items, expected ${queue.length}`);
            return false;
        }
        (_f = this.log) === null || _f === void 0 ? void 0 : _f.info(`Cache was broken, now fixed`);
        return true;
    }
}
exports.RemoteOCABundleResolver = RemoteOCABundleResolver;
